import json as jsonlib
import warnings
from datetime import datetime
from importlib.metadata import version as pkg_version
from typing import Any, Collection, Dict, List, Literal, Optional, Union
from urllib.error import HTTPError
from urllib.parse import urljoin

import requests
from packaging import version

from satcat.cli import utils as cli_utils
from satcat.sdk.events.event_client import EventClient
from satcat.sdk.od_client import ODClient
from satcat.sdk.propagation.propagation_client import PropagationClient
from satcat.sdk.screening.screening_client import ScreeningClient
from satcat.sdk.settings import settings
from satcat.sdk.utils import SortDirection

# Get current installed version
CURRENT_VERSION = pkg_version("satcat")


# Get latest version from PyPI
def get_latest_version(package_name: str) -> str:
    url = f"https://pypi.org/pypi/{package_name}/json"
    res = requests.get(url, timeout=2)
    if res.status_code == 200:
        return res.json()["info"]["version"]
    return None


# Compare versions and print a warning if outdated
def check_version():
    package_name = "satcat"
    try:
        current = version.parse(CURRENT_VERSION)
        latest = version.parse(get_latest_version(package_name))
        if latest and current < latest:
            warnings.warn(
                f"⚠️ You are using {package_name} {current}, "
                f"but the latest version is {latest}. "
                f"Run `pip install -U {package_name}` to upgrade."
            )
    except Exception:
        # Silently fail if check fails
        pass


AUTH_TIMEOUT_MINUTES = 60


class Client:
    """Base Client class which may be used to make requests to Satcat APIs.
    May be used as a context manager with the ``with`` keyword:

    .. code-block:: python

        with Client() as client:
            client.request(...)

    If not used with the ``with`` keyword, it is recommended to explicitly close
    the Client with ``client.close()`` after the application is finished making requests.
    """

    session: requests.Session

    def __init__(self, default_page_size: int = 32):
        """
        :param default_page_size: Default page size to use when sending repeated listing
            requests to a collection endpoint, defaults to 10
        :type page_size: int, optional
        """
        self.session = requests.Session()
        self.access_token = None
        self.authenticated = False
        self.authenticated_at = None
        self.default_page_size = default_page_size
        check_version()

    def __enter__(self):
        return self

    def __exit__(self, *exc_args):
        self.close()

    def close(self):
        """Close the client and any associated network connections."""
        self.session.close()

    def authenticate_session(self, timeout_s: Optional[int] = None):
        """Authenticate the HTTP session with the authentication provider using the
            authentication method determined by ``satcat.sdk.settings.auth_method`` .

        This is done automatically when using ``request_authenticated`` .
        """
        if settings.auth_method == "password":
            path = urljoin(settings.satcat_rest_api_url, "login")
            res = self.request(
                path,
                method="POST",
                data={
                    "username": settings.auth_username,
                    "password": settings.auth_password,
                    "grant_type": "password",
                },
            )
            try:
                token = res.json()["access_token"]
            except Exception:
                raise ValueError(
                    f"Authentication Error ({res.status_code}): {res.text}"
                ) from None
            self.access_token = token
            self.session.headers["Authorization"] = f"Bearer {token}"
        elif settings.auth_method == "client_credentials":
            payload = f"grant_type=client_credentials&client_id={settings.auth_client_id}&client_secret={settings.auth_client_secret}"
            headers = {"content-type": "application/x-www-form-urlencoded"}
            response = requests.post(
                urljoin(settings.satcat_rest_api_url, "/oauth/token/"),
                data=payload,
                headers=headers,
                timeout=timeout_s or settings.default_timeout,
            )

            if response.status_code == 200:
                token = response.json()["access_token"]
            else:
                raise ValueError(
                    "Failed to authenticate with client_credentials flow using provided credentials"
                )
            self.session.headers["Authorization"] = f"Bearer {token}"
            self.access_token = token
        self.authenticated = True
        self.authenticated_at = datetime.utcnow()

    def request(
        self,
        path: str,
        method: str = "GET",
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        params: Optional[dict] = None,
        files=None,
        api: Literal["rest", "web"] = "rest",
    ) -> requests.Response:
        """Make an HTTP request to a Satcat API.
        .. warning:: For most use cases, this function should only be called internally by
        high level functions inside the SDK, such as those inside ``client.screening`` .
        """
        if api == "rest":
            url = urljoin(settings.satcat_rest_api_url, path)
        elif api == "web":
            url = urljoin(settings.satcat_web_api_url, path)
        else:
            raise ValueError(f"Invalid api setting {api}")

        with cli_utils.FetchingSpinner():
            response = self.session.request(
                method,
                url,
                json=json,
                data=data,
                params=params,
                files=files,
                timeout=settings.default_timeout,
            )
        return response

    def request_authenticated(
        self,
        path: str,
        method: str = "GET",
        data: Optional[dict] = None,
        json: Optional[dict] = None,
        params: Optional[dict] = None,
        ignore_failure: bool = False,
        files=None,
        api: Literal["rest", "web"] = "rest",
    ) -> requests.Response:
        """Ensure that the HTTP session is authorized and make an HTTP request to
            a Satcat API.

        .. warning:: For most use cases, this function should only be called internally by
            high level functions inside the SDK, such as those inside ``client.screening`` .
        """
        if not self.authenticated:
            with cli_utils.FetchingSpinner("Authenticating..."):
                self.authenticate_session()
        elif (
            self.authenticated_at
            and (datetime.utcnow() - self.authenticated_at).total_seconds()
            > AUTH_TIMEOUT_MINUTES * 60
        ):
            with cli_utils.FetchingSpinner("Authenticating..."):
                self.authenticate_session()
        res = self.request(
            path, method, data=data, json=json, params=params, files=files, api=api
        )
        if not str(res.status_code).startswith("2") and not ignore_failure:
            raise HTTPError(path, res.status_code, res.text, res.headers, None)
        return res

    def request_list(
        self,
        path: str,
        filters: Optional[List[Dict]] = None,
        count: Optional[int] = None,
        params: Optional[dict] = None,
        sort_direction: Union[SortDirection, str] = SortDirection.DESC,
        sort_field: str = "created_at",
        fields: Optional[List[str]] = None,
    ) -> Collection[Any]:
        """Request a list of resources from a Satcat API and automatically page through
            the results if necessary.

        .. warning:: For most use cases, this function should only be called internally by
            high level functions inside the SDK, such as those inside ``client.screening`` .
        """
        if params is None:
            params = {}
        params["sort_field"] = sort_field
        if fields is not None:
            params["fields"] = ",".join(fields)
        if isinstance(sort_direction, str):
            sort_direction = SortDirection(sort_direction)
        params["sort_direction"] = sort_direction.value.lower()
        result_items = []
        if filters is not None:
            params["filters"] = jsonlib.dumps(filters)
        if count is not None:
            params["count"] = count
            result_items = self.request_authenticated(
                path, "GET", params=params
            ).json()["items"]
        else:
            params["count"] = self.default_page_size
            result_items = []
            offset = 0
            total_count = None
            while total_count is None or offset <= total_count:
                params["offset"] = offset
                res = self.request_authenticated(path, "GET", params=params)
                data = res.json()
                items = data["items"]
                result_items.extend(items)
                offset += self.default_page_size
                total_count = data["total_count"]

        for item in result_items:
            item["client"] = self

        return result_items

    def ping_authenticated(self) -> bool:
        """Sends a request to Kayhan servers to confirm a healthy and authenticated session. Returns True if valid, raises an HTTPError if not."""
        try:
            self.request_authenticated("health_authenticated")
            return True
        except HTTPError as e:
            if e.code == 401:
                return False
            raise

    @property
    def propagation(self) -> "PropagationClient":
        """Library containing SDK functions for interacting with the
        Satcat Propagation API."""
        return PropagationClient(self)

    @property
    def screening(self) -> "ScreeningClient":
        """Library containing SDK functions for interacting with the
        Satcat Screening API."""
        return ScreeningClient(self)

    @property
    def events(self) -> "EventClient":
        """Library containing SDK functions for interacting with the
        Satcat Events API."""
        return EventClient(self)

    @property
    def od(self) -> "ODClient":
        """Library containing SDK functions for interacting with the
        Satcat OD API."""
        return ODClient(self)

    def create_ephemeris(self, *args, **kwargs):
        """Alias for ``client.screening.create_ephemeris``."""
        return self.screening.create_ephemeris(*args, **kwargs)

    def list_ephemerides(self, *args, **kwargs):
        """Alias for ``client.screening.list_ephemerides``."""
        return self.screening.list_ephemerides(*args, **kwargs)

    def get_ephemeris(self, *args, **kwargs):
        """Alias for ``client.screening.get_ephemeris``."""
        return self.screening.get_ephemeris(*args, **kwargs)

    def download_ephemeris(self, *args, **kwargs):
        """Alias for ``client.screening.download_ephemeris``."""
        return self.screening.download_ephemeris(*args, **kwargs)

    def operationalize_ephemeris(self, *args, **kwargs):
        """Alias for ``client.screening.operationalize_ephemeris``."""
        return self.screening.operationalize_ephemeris(*args, **kwargs)



import heapq
import os
from datetime import datetime
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, TextIO, Union
from urllib.parse import urljoin

from satcat.sdk.settings import settings
from satcat.sdk.utils import PathLike, open_path_or_buf

if TYPE_CHECKING:
    from satcat.sdk.client import Client


class ODClient:
    client: "Client"

    def __init__(self, client: "Client"):
        self.client = client

    def list_assets(self) -> List[int]:
        """
        List assets for which the user has orbital determination data available.

        :return The list of object IDs.
        :rtype: Collection[int]
        """

        path = urljoin(settings.satcat_rest_api_url, "od/objects")
        items = self.client.request_authenticated(path)
        return items.json()["objects"]

    def list_object_files(
        self, object_id: int, today: bool = False, results: bool = False
    ) -> Dict[str, Any]:
        """
        List orbital determination files for a specific object.

        :param object_id: The NORAD ID or temporary ID of the object.
        :type object_id: int
        :param today: If True, only list files uploaded today.
        :type today: bool
        :param results: If True, only list result files.
        :type results: bool
        :return: The list of files for the specified object.
        :rtype: Collection[Dict[str, Any]]
        """

        path = urljoin(settings.satcat_rest_api_url, f"od/{object_id}")
        params = {"today": str(today).lower(), "results": str(results).lower()}
        items = self.client.request_authenticated(path, params=params)
        return items.json()["files"]

    def upload_file(
        self,
        object_id: int,
        path_or_buf: Union[PathLike, TextIO, StringIO],
        filename: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Upload an orbital determination (OD) file for a specific object.

        The file content can be provided either as a filesystem path or as an
        already-open file-like object. A filename must either be provided
        explicitly or derivable from ``path_or_buf``; otherwise this method
        will raise a ``ValueError``.

        :param object_id: The NORAD ID or temporary ID of the object that this
            OD file describes.
        :type object_id: int
        :param path_or_buf: Path to a file on disk, or an open text file-like
            object (e.g. ``io.StringIO``) containing the OD data to upload.
        :type path_or_buf: Union[PathLike, TextIO, StringIO]
        :param filename: Optional explicit filename to send to the API. If not
            provided, the filename will be derived from ``path_or_buf`` when
            possible (e.g. using ``os.path.basename``); if no filename can be
            determined, a ``ValueError`` is raised.
        :type filename: Optional[str]
        :return: The JSON response from the OD upload endpoint.
        :rtype: Dict[str, Any]
        """

        with open_path_or_buf(path_or_buf, mode="rb") as file:
            resolved_filename = filename

            if resolved_filename is None:
                # Try to derive the filename from the original input
                if isinstance(path_or_buf, Path):
                    resolved_filename = os.path.basename(path_or_buf.name)
                elif isinstance(path_or_buf, str):
                    resolved_filename = os.path.basename(path_or_buf)
                else:
                    # Fall back to any name attribute on the underlying file object
                    file_name_attr = getattr(file, "name", None)
                    if isinstance(file_name_attr, str):
                        resolved_filename = os.path.basename(file_name_attr)

            if not resolved_filename:
                raise ValueError(
                    "A filename must be provided explicitly or derivable from path_or_buf."
                )

            path = urljoin(settings.satcat_rest_api_url, f"od/{object_id}/upload")
            files = {"file": (resolved_filename, file)}
            response = self.client.request_authenticated(
                path, method="POST", files=files
            )
        return response.json()

    def download_file(
        self, object_id: int, filename: str, output_path: Optional[PathLike]
    ) -> bytes:
        """
        Download an orbital determination (OD) file for a specific object.

        :param object_id: The NORAD ID or temporary ID of the object that this
            OD file describes.
        :type object_id: int
        :param filename: The name of the file to download.
        :type filename: str
        :param output_path: Optional path to save the downloaded file. If None, the file content is returned as bytes.
        :type output_path: Optional[PathLike]
        :return: The content of the downloaded file as bytes.
        :rtype: bytes
        """

        path = urljoin(settings.satcat_rest_api_url, f"od/{object_id}/file")
        response = self.client.request_authenticated(
            path, method="GET", json={"file_name": filename}
        )
        if output_path:
            expanded_path = os.path.abspath(os.path.expanduser(str(output_path)))
            os.makedirs(os.path.dirname(expanded_path), exist_ok=True)
            with open(expanded_path, "wb") as outf:
                outf.write(response.content)
        return response.content

    def download_recent_files(
        self,
        object_id: int,
        number: int = 1,
        report: bool = False,
        output_dir: Optional[PathLike] = None,
    ):
        """
        Download the most recent orbital determination (OD) files for a specific object.

        :param object_id: The NORAD ID or temporary ID of the object that this
            OD file describes.
        :type object_id: int
        :param number: The number of recent files to download.
        :type number: int
        :param report: If True, downloads the most recent pdf report file as well
        :type report: bool
        :param output_dir: Optional directory to save the downloaded files. If None, files are saved to the current working directory.
        :type output_dir: Optional[PathLike]
        :return: A tuple containing the count of definitive files, predictive files, and a boolean indicating if a PDF report was downloaded.
        :rtype: Tuple[int, int, bool]
        """

        files = self.list_object_files(object_id, results=True)

        file_heap = []

        definitive_count = 0
        predictive_count = 0
        pdf_downloaded = False

        for file_info in files:
            filename = file_info["filename"]
            if (
                "DEFINITIVE" in filename
                or "PREDICTIVE" in filename
                or filename.endswith(".pdf")
            ):
                heapq.heappush(
                    file_heap,
                    (
                        -datetime.fromisoformat(file_info["last_modified"]).timestamp(),
                        filename,
                    ),
                )

        # Create or set output directory
        if output_dir:
            expanded_dir = os.path.abspath(os.path.expanduser(str(output_dir)))
            os.makedirs(expanded_dir, exist_ok=True)
            path_base = expanded_dir
        else:
            path_base = os.getcwd()

        while file_heap:
            _, file_key = heapq.heappop(file_heap)
            file_name = os.path.basename(file_key)
            output_path = os.path.join(path_base, file_name)

            if (
                definitive_count == number
                and predictive_count == number
                and (pdf_downloaded or not report)
            ):
                break

            if file_name.endswith(".txt"):
                if "DEFINITIVE" in file_name and definitive_count < number:
                    self.download_file(object_id, file_key, output_path=output_path)
                    definitive_count += 1
                elif "PREDICTIVE" in file_name and predictive_count < number:
                    self.download_file(object_id, file_key, output_path=output_path)
                    predictive_count += 1
            elif report and file_name.endswith(".pdf") and not pdf_downloaded:
                self.download_file(object_id, file_key, output_path=output_path)
                pdf_downloaded = True

        return definitive_count, predictive_count, pdf_downloaded
